use std::collections::HashMap;
use std::path::PathBuf;
use sv_parser::{parse_sv, unwrap_node, ConstantMintypmaxExpression, GateInstantiation, ListOfParameterAssignments, ListOfPortConnections, Locate, PackedDimensionRange, RefNode, SyntaxTree};
use crate::core::fast_hdlparam::FastHdlparam;

pub fn sv_parser(path: &str) -> FastHdlparam {
    // The path of SystemVerilog source file
    let path = PathBuf::from(path);
    // The list of defined macros
    let defines = HashMap::new();
    // The list of include paths
    let includes: Vec<PathBuf> = Vec::new();

    // Parse
    let result: Result<(SyntaxTree, HashMap<String, Option<sv_parser::Define>>), sv_parser::Error> = parse_sv(&path, &defines, &includes, false, true);

    let hdlparam = make_fast_from_syntaxtree(result);

    hdlparam
}

pub fn make_fast_from_syntaxtree(result: Result<(SyntaxTree, HashMap<String, Option<sv_parser::Define>>), sv_parser::Error>) -> FastHdlparam {
    let mut hdlparam = FastHdlparam {
        fast: Vec::new()
    };

    let mut nonansi_port_locate = HashMap::new();
    let mut ansi_port_last_dir = "";

    if let Ok((syntax_tree, _)) = result {
        let content = syntax_tree.text.text().split('\n')
                                                    .map(|s| s.to_string())
                                                    .collect::<Vec<String>>();
        // &SyntaxTree is iterable
        for node in &syntax_tree {
            match node {
                RefNode::ModuleDeclaration(x) => {
                    let id = unwrap_node!(x, ModuleIdentifier).unwrap();
                    let id = get_identifier(id).unwrap();
                    let (line, character) = (id.line, get_column_by_offset(&content, id.offset) as u32);
                    let end_character = character + id.len as u32;
                    let name = syntax_tree.get_str(&id).unwrap();
                    hdlparam.new_module(name, line, character, end_character);
                }
                RefNode::ParameterPortDeclaration(x) => {
                    let id = unwrap_node!(x, ParameterIdentifier).unwrap();
                    let id = get_identifier(id).unwrap();
                    let name = syntax_tree.get_str(&id).unwrap();
                    match unwrap_node!(x, ParameterDeclarationParam, ParameterPortDeclarationParamList) {
                        Some(RefNode::ParameterDeclarationParam(param_node)) => {
                            // println!("{:?}", param_node);
                            let keyword_locate = param_node.nodes.0.nodes.0;
                            // println!("keyword {:#?}", keyword_locate);
                            let (start_line, start_character) = (id.line, get_column_by_offset(&content, keyword_locate.offset) as u32);
                            let (mut end_line, mut end_character) = (id.line, start_character + id.len as u32);
                            let net_type = match unwrap_node!(param_node, DataType) {
                                Some(RefNode::DataType(data_type)) => {
                                    let id = get_identifier(unwrap_node!(data_type, SimpleIdentifier, Keyword).unwrap()).unwrap();
                                    syntax_tree.get_str(&id).unwrap()
                                }
                                _ => "wire"
                            };
                            let init = match unwrap_node!(param_node, ConstantMintypmaxExpression).unwrap() {
                                RefNode::ConstantMintypmaxExpression(expression) => {
                                    // println!("expression {:?}", expression);
                                    let (exp, last_locate) = parse_parameter_expression(&syntax_tree, expression);
                                    (end_line, end_character) = if last_locate != None {
                                        // println!("param {:?} lastlocate {:?}", name, last_locate);
                                        (last_locate.unwrap().line, (get_column_by_offset(&content, last_locate.unwrap().offset) + last_locate.unwrap().len) as u32)
                                    } else { 
                                        (end_line, end_character) 
                                    };
                                    // println!("end pos {} {}", end_line, end_character);
                                    exp
                                }
                                _ => "unknown".to_string()
                            };
                            hdlparam.add_parameter(name, net_type, init.as_str(), start_line, start_character, end_line, end_character);
                        }
                        _ => ()
                    }
                }
                RefNode::Port(x) => {
                    let id = unwrap_node!(x, SimpleIdentifier).unwrap();
                    let locate = get_identifier(id).unwrap();
                    let name = syntax_tree.get_str(&locate.clone()).unwrap();
                    // println!("get port {} {:?}", name, locate);
                    nonansi_port_locate.insert(name, locate);
                }
                RefNode::PortDeclaration(x) => {
                    let id = unwrap_node!(x, InputDeclaration, OutputDeclaration, InoutDeclaration).unwrap();
                    let id = get_identifier(id).unwrap();
                    let dir_type = syntax_tree.get_str(&id).unwrap();
                    let (dir_line, dir_character) = (id.line, get_column_by_offset(&content, id.offset) as u32);

                    let net_type = match unwrap_node!(x, DataType, ImplicitDataType) {
                        Some(RefNode::DataType(x)) => {
                            let id = unwrap_node!(x, Keyword).unwrap();
                            syntax_tree.get_str(&get_identifier(id).unwrap()).unwrap()
                        },
                        Some(RefNode::ImplicitDataType(_)) => "wire",
                        _ => "unknown"
                    };

                    let width = match unwrap_node!(x, PackedDimensionRange) {
                        Some(RefNode::PackedDimensionRange(x)) => {
                            parse_port_expression(&syntax_tree, x)
                        }
                        _ => "1".to_string()
                    };

                    let id = unwrap_node!(x, PortIdentifier).unwrap();
                    let id = get_identifier(id).unwrap();
                    let name = syntax_tree.get_str(&id).unwrap();
                    let (start_line, start_character, end_line, end_character) = if nonansi_port_locate.contains_key(name) {
                        let start_character = get_column_by_offset(&content, nonansi_port_locate[name].offset);
                        (nonansi_port_locate[name].line, start_character as u32,
                            nonansi_port_locate[name].line, (start_character + nonansi_port_locate[name].len) as u32)
                    } else {
                        (dir_line, dir_character, id.line, (get_column_by_offset(&content, id.offset) + id.len) as u32)
                    };

                    hdlparam.add_port(name, dir_type, net_type, width.as_str(), start_line, start_character, end_line, end_character);
                }
                RefNode::AnsiPortDeclaration(x) => {
                    let id = unwrap_node!(x, PortIdentifier).unwrap();
                    let name_locate = get_identifier(id).unwrap();
                    let name = syntax_tree.get_str(&name_locate).unwrap();
                    let character = get_column_by_offset(&content, name_locate.offset);
                    let (end_line, end_character) = (name_locate.line, (character + name_locate.len) as u32);

                    let id = unwrap_node!(x, PortDirection);
                    let (start_line, start_character) = if id != None {
                        let id = id.unwrap();
                        let dir_locate = get_identifier(id).unwrap();
                        ansi_port_last_dir = syntax_tree.get_str(&dir_locate).unwrap();
                        (dir_locate.line, get_column_by_offset(&content, dir_locate.offset) as u32)
                    } else {
                        (name_locate.line, character as u32)
                    };

                    let net_type = if unwrap_node!(x, AnsiPortDeclarationVariable) != None {
                        "wire"
                    } else {
                        match unwrap_node!(x, DataType, ImplicitDataType) {
                            Some(RefNode::DataType(x)) => {
                                let id = unwrap_node!(x, Keyword).unwrap();
                                syntax_tree.get_str(&get_identifier(id).unwrap()).unwrap()
                            },
                            Some(RefNode::ImplicitDataType(_)) => "wire",
                            _ => "unknown"
                        }
                    };

                    let width = match unwrap_node!(x, PackedDimensionRange) {
                        Some(RefNode::PackedDimensionRange(x)) => {
                            parse_port_expression(&syntax_tree, x)
                        }
                        _ => "1".to_string()
                    };

                    hdlparam.add_port(name, ansi_port_last_dir, net_type, width.as_str(), start_line, start_character, end_line, end_character);
                }
                RefNode::ModuleInstantiation(x) => {
                    let id = unwrap_node!(x, ModuleIdentifier).unwrap();
                    let id = get_identifier(id).unwrap();
                    let inst_type = syntax_tree.get_str(&id).unwrap();

                    let id = unwrap_node!(x, HierarchicalInstance).unwrap();
                    let id = get_identifier(id).unwrap();
                    let name = syntax_tree.get_str(&id).unwrap();
                    let (line, character) = (id.line, get_column_by_offset(&content, id.offset) as u32);
                    let end_character = character + id.len as u32;

                    let (param_start_line, param_start_character,
                        param_end_line, param_end_character) = match unwrap_node!(x, ListOfParameterAssignments) {
                        Some(RefNode::ListOfParameterAssignments(x)) => {
                            let locate = get_inst_param_last_locate(x);
                            if locate.is_none() {
                                (0, 0, 0, 0)
                            } else {
                                (
                                    locate.unwrap().0.line, get_column_by_offset(&content, locate.unwrap().0.offset) as u32,
                                    locate.unwrap().1.line, (get_column_by_offset(&content, locate.unwrap().1.offset) + locate.unwrap().1.len) as u32
                                )
                            }
                        }
                        _ => (0, 0, 0, 0)
                    };

                    let (port_start_line, port_start_character,
                        port_end_line, port_end_character) = match unwrap_node!(x, ListOfPortConnections) {
                        Some(RefNode::ListOfPortConnections(x)) => {
                            let locate = get_inst_port_last_locate(x);
                            if locate.is_none() {
                                (0, 0, 0, 0)
                            } else {
                                (
                                    locate.unwrap().0.line, get_column_by_offset(&content, locate.unwrap().0.offset) as u32,
                                    locate.unwrap().1.line, (get_column_by_offset(&content, locate.unwrap().1.offset) + locate.unwrap().1.len) as u32
                                )
                            }
                        }
                        _ => (0, 0, 0, 0)
                    };

                    hdlparam.add_instance(name, inst_type, line, character, end_character, 
                        param_start_line, param_start_character, param_end_line, param_end_character, 
                        port_start_line, port_start_character, port_end_line, port_end_character
                    );
                }
                RefNode::GateInstantiation(x) => {
                    let id = unwrap_node!(x, GateInstantiation).unwrap();
                    let id = get_identifier(id).unwrap();
                    let inst_type = syntax_tree.get_str(&id).unwrap();

                    let id = unwrap_node!(x, NInputGateInstance, NOutputGateInstance).unwrap();
                    let id = get_identifier(id).unwrap();
                    let name = syntax_tree.get_str(&id).unwrap();
                    let (line, character) = (id.line, get_column_by_offset(&content, id.offset) as u32);
                    let end_character = character + id.len as u32;

                    let locate = get_gateinst_port_locate(x);
                    let (param_start_line, param_start_character, param_end_line, param_end_character) = (0, 0, 0, 0);
                    let (port_start_line, port_start_character, port_end_line, port_end_character) = if locate.is_none() {
                        (0,0,0,0)
                    } else {
                        (
                            locate.unwrap().0.line, get_column_by_offset(&content, locate.unwrap().0.offset) as u32,
                            locate.unwrap().1.line, (get_column_by_offset(&content, locate.unwrap().1.offset) + locate.unwrap().1.len) as u32
                        )
                    };

                    hdlparam.add_instance(name, inst_type, line, character, end_character, 
                        param_start_line, param_start_character, param_end_line, param_end_character, 
                        port_start_line, port_start_character, port_end_line, port_end_character
                    );
                }
                _ =>  ()
            }
        }
    } else {
        println!("Parse failed");
    }

    hdlparam
}

fn get_inst_param_last_locate(x: &ListOfParameterAssignments) -> Option<(Locate, Locate)> {
    let mut first_locate = Locate { offset: 0, line: 0, len: 0 };
    let mut last_locate = Locate { offset: 0, line: 0, len: 0 };

    for node in x {
        match unwrap_node!(node, SimpleIdentifier, Symbol) {
            Some(RefNode::SimpleIdentifier(x)) => {
                let locate = x.nodes.0;
                if locate != last_locate { last_locate = locate; }
                if first_locate.offset == 0 { first_locate = locate; };
            }
            Some(RefNode::Symbol(x)) => {
                let locate = x.nodes.0;
                if locate != last_locate { last_locate = locate; }
                if first_locate.offset == 0 { first_locate = locate; };
            }
            _ => ()
        }
    }

    if last_locate.offset == 0 {
        None
    } else {
        Some((first_locate, last_locate))
    }
}

fn get_inst_port_last_locate(x: &ListOfPortConnections) -> Option<(Locate, Locate)> {
    let mut first_locate = Locate { offset: 0, line: 0, len: 0 };
    let mut last_locate = Locate { offset: 0, line: 0, len: 0 };

    for node in x {
        match unwrap_node!(node, SimpleIdentifier, Symbol) {
            Some(RefNode::SimpleIdentifier(x)) => {
                let locate = x.nodes.0;
                if locate != last_locate { last_locate = locate; }
                if first_locate.offset == 0 { first_locate = locate; };
            }
            Some(RefNode::Symbol(x)) => {
                let locate = x.nodes.0;
                if locate != last_locate { last_locate = locate; }
                if first_locate.offset == 0 { first_locate = locate; };
            }
            _ => ()
        }
    }

    if last_locate.offset == 0 {
        None
    } else {
        Some((first_locate, last_locate))
    }
}

fn get_gateinst_port_locate(x: &GateInstantiation) -> Option<(Locate, Locate)> {
    let mut first_locate = Locate { offset: 0, line: 0, len: 0 };
    let mut last_locate = Locate { offset: 0, line: 0, len: 0 };

    for node in x {
        match unwrap_node!(node, InputTerminal, OutputTerminal) {
            Some(RefNode::InputTerminal(x)) => {
                let id = unwrap_node!(x, Identifier).unwrap();
                let locate = get_identifier(id).unwrap();
                if locate != last_locate { last_locate = locate; }
                if first_locate.offset == 0 { first_locate = locate; };
            }
            Some(RefNode::OutputTerminal(x)) => {
                let id = unwrap_node!(x, Identifier).unwrap();
                let locate = get_identifier(id).unwrap();
                if locate != last_locate {
                    last_locate = locate;
                }
            }
            _ => ()
        }
    }

    if last_locate.offset == 0 {
        None
    } else {
        Some((first_locate, last_locate))
    }
}

fn parse_port_expression(syntax_tree: &SyntaxTree, x: &PackedDimensionRange) -> String {
    let mut last_locate = Locate { offset: 0, line: 0, len: 0 };
    let mut expression = String::new();
    for node in x {
        // println!("parse expression::node {:#?}", node);
        match unwrap_node!(node, SimpleIdentifier, Symbol, UnsignedNumber, HexNumber, OctalNumber, BinaryNumber) {
            Some(RefNode::SimpleIdentifier(x)) => {
                let locate = x.nodes.0;
                if locate != last_locate {
                    last_locate = locate;
                    let s = syntax_tree.get_str(&locate).unwrap();
                    expression = expression + s;
                    // println!("parse expression {}", s);
                }
            }
            Some(RefNode::Symbol(x)) => {
                let locate = x.nodes.0;
                if locate != last_locate {
                    last_locate = locate;
                    let s = syntax_tree.get_str(&x.nodes.0).unwrap();
                    expression = expression + s;
                    // println!("parse expression {}", s);
                }
            }
            Some(RefNode::UnsignedNumber(x)) => {
                let locate = x.nodes.0;
                if locate != last_locate {
                    last_locate = locate;
                    let s = syntax_tree.get_str(&x.nodes.0).unwrap();
                    expression = expression + s;
                    // println!("parse expression {}", s);
                }
            }
            Some(RefNode::HexNumber(x)) => {
                let locate = x.nodes.1.nodes.0;
                if locate != last_locate {
                    last_locate = locate;
                    let size = if x.nodes.0 != None { syntax_tree.get_str(&x.nodes.0).unwrap() } else { "" };
                    let base = syntax_tree.get_str(&x.nodes.1.nodes.0).unwrap();
                    let number = syntax_tree.get_str(&x.nodes.2.nodes.0).unwrap();
                    expression = expression + size + base + number;
                    // println!("parse expression {}", expression);
                }
            }
            Some(RefNode::OctalNumber(x)) => {
                let locate = x.nodes.1.nodes.0;
                if locate != last_locate {
                    last_locate = locate;
                    let size = if x.nodes.0 != None { syntax_tree.get_str(&x.nodes.0).unwrap() } else { "" };
                    let base = syntax_tree.get_str(&x.nodes.1.nodes.0).unwrap();
                    let number = syntax_tree.get_str(&x.nodes.2.nodes.0).unwrap();
                    expression = expression + size + base + number;
                    // println!("parse expression {}", expression);
                }
            }
            Some(RefNode::BinaryNumber(x)) => {
                let locate = x.nodes.1.nodes.0;
                if locate != last_locate {
                    last_locate = locate;
                    let size = if x.nodes.0 != None { syntax_tree.get_str(&x.nodes.0).unwrap() } else { "" };
                    let base = syntax_tree.get_str(&x.nodes.1.nodes.0).unwrap();
                    let number = syntax_tree.get_str(&x.nodes.2.nodes.0).unwrap();
                    expression = expression + size + base + number;
                    // println!("parse expression {}", expression);
                }
            }
            _ => ()
        }
    }
    if expression == "" {
        "unknown".to_string()
    } else {
        // println!("get {}", expression);
        expression
    }
}

fn parse_parameter_expression(syntax_tree: &SyntaxTree, x: &ConstantMintypmaxExpression) -> (String, Option<Locate>) {
    let mut last_locate = Locate { offset: 0, line: 0, len: 0 };
    let mut expression = String::new();
    for node in x {
        // println!("parse expression::node {:#?}", node);
        match unwrap_node!(node, SimpleIdentifier, Symbol, UnsignedNumber, HexNumber, OctalNumber, BinaryNumber) {
            Some(RefNode::SimpleIdentifier(x)) => {
                let locate = x.nodes.0;
                if locate != last_locate {
                    last_locate = locate;
                    let s = syntax_tree.get_str(&locate).unwrap();
                    expression = expression + s;
                    // println!("parse expression {}", s);
                }
            }
            Some(RefNode::Symbol(x)) => {
                let locate = x.nodes.0;
                if locate != last_locate {
                    last_locate = locate;
                    let s = syntax_tree.get_str(&x.nodes.0).unwrap();
                    expression = expression + s;
                    // println!("parse expression {}", s);
                }
            }
            Some(RefNode::UnsignedNumber(x)) => {
                let locate = x.nodes.0;
                if locate != last_locate {
                    last_locate = locate;
                    let s = syntax_tree.get_str(&x.nodes.0).unwrap();
                    expression = expression + s;
                    // println!("parse expression {}", s);
                }
            }
            Some(RefNode::HexNumber(x)) => {
                let locate = x.nodes.1.nodes.0;
                if locate != last_locate {
                    last_locate = locate;
                    let size = if x.nodes.0 != None { syntax_tree.get_str(&x.nodes.0).unwrap() } else { "" };
                    let base = syntax_tree.get_str(&x.nodes.1.nodes.0).unwrap();
                    let number = syntax_tree.get_str(&x.nodes.2.nodes.0).unwrap();
                    expression = expression + size + base + number;
                    // println!("parse expression {}", expression);
                }
            }
            Some(RefNode::OctalNumber(x)) => {
                let locate = x.nodes.1.nodes.0;
                if locate != last_locate {
                    last_locate = locate;
                    let size = if x.nodes.0 != None { syntax_tree.get_str(&x.nodes.0).unwrap() } else { "" };
                    let base = syntax_tree.get_str(&x.nodes.1.nodes.0).unwrap();
                    let number = syntax_tree.get_str(&x.nodes.2.nodes.0).unwrap();
                    expression = expression + size + base + number;
                    // println!("parse expression {}", expression);
                }
            }
            Some(RefNode::BinaryNumber(x)) => {
                let locate = x.nodes.1.nodes.0;
                if locate != last_locate {
                    last_locate = locate;
                    let size = if x.nodes.0 != None { syntax_tree.get_str(&x.nodes.0).unwrap() } else { "" };
                    let base = syntax_tree.get_str(&x.nodes.1.nodes.0).unwrap();
                    let number = syntax_tree.get_str(&x.nodes.2.nodes.0).unwrap();
                    expression = expression + size + base + number;
                    // println!("parse expression {}", expression);
                }
            }
            _ => ()
        }
    }
    if expression == "" {
        ("unknown".to_string(), None)
    } else {
        // println!("parse function lastlocate {:?}", last_locate);
        (expression, Some(last_locate))
    }
}

fn get_identifier(node: RefNode) -> Option<Locate> {
    // unwrap_node! can take multiple types
    match unwrap_node!(node, SimpleIdentifier, EscapedIdentifier, Keyword) {
        Some(RefNode::SimpleIdentifier(x)) => {
            return Some(x.nodes.0);
        }
        Some(RefNode::EscapedIdentifier(x)) => {
            return Some(x.nodes.0);
        }
        Some(RefNode::Keyword(x)) => {
            return Some(x.nodes.0);
        }
        _ => None,
    }
}

fn get_column_by_offset(content: &[String], offset: usize) -> usize {
    let mut current_offset = 0;

    for line_content in content {
        let line_length = line_content.len() + 1; // +1 for newline character
        // println!("now line {} len {}", line_content, line_length);
        // println!("now offset {} offset {}", current_offset + line_length, offset);
        if current_offset + line_length > offset {
            return offset - current_offset + 1;
        }
        current_offset += line_length;
    }

    1 // if offset over the file lentgh，return 1
}
